% This Matlab script demonstrates the Turbo AMP reconstruction
% for Compressive Imaging. 
% 
% turboAMP version 1.1 
% Copyright (c) Subhojit Som and Philip Schniter, 2013
% Email: som.4@osu.edu, schniter@ece.osu.edu
%
% Last modified: February 10, 2013
%
% New in version 1.1:
% 1) Support for un-normalized sensing matrix A
% 2) Support for arbitrary scaling of signal x
% 3) High and Low SNR modes
% 4) Support for function handles for A



close all; clear all; clear java; clc;


path(path,'utils/');
path(path,'hmt/');
path(path,'turbo/');


%%% read image file %%% 
% X=imread('cameraman.tif');
% X=X(21:148,71:198);
% X=double(X)/255;
% X=imresize(X,0.5,'bicubic');

load imageData.mat; % This directly loads image as variable X. May use
                    % the previous lines to load the image as well and
		            % comment this line.
M=5000;             % number of measurements



SNRdB_true=Inf;% SNR  in dB 
numLevels=4;   % wavelet tree depth 
A_type = 1;    % Measurement matrix: 1=iid Gaussian, 2=Rademacher


%%% settings for turbo AMP algorithm
params = getConfigTurboAMP();
params.projOnImage=1;  % 0: proj on wavelet trandsform, 1: proj on image

% % uncomment the following line for low SNR
% params.highSNR=false;  % whether the observation has high SNR 

[A,X,M] = getA(X,img,M,A_type);


dim=size(X);  % dimension of the image
N=prod(dim);  % total number of pixels


for n=1:N
    A(:,n) = A(:,n)/norm(A(:,n)); 
end

%%% wavelet trandform of the image %%%
[waveC, signature]=wavedec2(X,numLevels,'db1');
W=wave_vec2mat(waveC,signature);
params.numLevels=numLevels;
params.s=signature;


if(params.projOnImage) 
    x_true=X(:);
else
    x_true=W(:);
end
y = A*x_true;

sig2w_true=norm(y)^2/M*10^(-SNRdB_true/10);     
w=sqrt(sig2w_true)*randn(M,1);
y=y+w;


%%% Model Parameters %%%
% [modelParams, ~]=initializeModelParams(params);

% The following parameters can be set by the user. If nothing is
% set the default values mentioned in the paper are used.
%
% modelParams.Gamma=[1 1e-6]; % noise Gamma params [a b]
% modelParams.Gamma_j= ...;    % [(numLevels+1)x2, [a_j b_j] per row, starting at j=-1]
% modelParams.Beta_0_1= ...;   % [1x2 double, [c d]]
% modelParams.Beta_j_11= ...;  % [(numLevels-1)x2 [c_j d_j] per row, starting at j=0]
% modelParams.Beta_j_00= ...;  % [(numLevels-1)x2 [\underbar{c}_j \underbar{d}_j] per row, starting at j=0]
% modelParams.Beta_approx= ...;% [1x2, [\underbar{c} \underbar{d}]]


% Natural image pixels are typically in the range [0, xRange].
% If the image is normalized, use xRange=1. If the image 
% is in the range [0, 255] use xRange=255.
xRange=1;
% xRange=255; 

% matrix form - default model parameters
[x_hat, s_hat]=turboAMP(y,A,[],params,xRange);

% % matrix form - user defined model parameters 
% [x_hat, s_hat]=turboAMP(y,A,[],params,xRange,modelParams);

% % function handle form - default model parameters
% B = @(z) A*z;
% Bt = @(z) A'*z;
% [x_hat, s_hat]=turboAMP(y,B,Bt,params,xRange);

% % function handle form - user defined model parameters
% B = @(z) A*z;
% Bt = @(z) A'*z;
% [x_hat, s_hat]=turboAMP(y,B,Bt,params,xRange,modelParams);


if(params.projOnImage)
    figure(1);
    % if you do not have image processing toolbox, comment the following
    % two lines and uncomment the next two lines
    subplot(121);imshow(reshape(x_true,dim));colormap('gray');title('original image');axis square;
    subplot(122);imshow(reshape(x_hat,dim));colormap('gray');title('reconstructed image');axis square;
%     subplot(121);imagesc(reshape(x_true,dim));colormap('gray');title('original image');axis square;
%     subplot(122);imagesc(reshape(x_hat,dim));colormap('gray');title('reconstructed image');axis square;
else
    figure;
    subplot(211);plot(x_true);title('original wavelet coefficients');
    subplot(212);plot(x_hat,'r');title('reconstructed wavelet coefficients');
end




